/*	File:  ServoMain.c
Firmware for Closed Loop Servo Controller

Adapted from code by Virgilio Mattoli, v 3.1.2, Feb 6, 2006
40MHz clock (10MHz*4 via PLL)

Jonathan Karpick
May 1, 2006


*/

///////////////////
// HEADER FILES
///////////////////
#include "ServoMain.h"

///////////////////
// MODULE VARIABLES
///////////////////
static unsigned char mTempCounter;		// Holder for DelayCounter1 during interrupt

///////////////////
// PRIVATE FUNCTION PROTOTYPES
///////////////////

void highISR( void );
void lowISR( void );
void TimerUpdate( void );

////////////////////////
// Function:	main
void main( void ) {
	
	unsigned int i;			// debugging

	SetPorts();
	VarInit();
	InitI2C();
	InitSensors();

	G_LED = 1;				//Hello Start
	MILLISECONDS (250);
	MILLISECONDS (250);
	G_LED = 0;				//Hello Middle
	R_LED = 1;				//Hello Middle
	MILLISECONDS (250);
	MILLISECONDS (250);
	R_LED = 0;				//Hello Last
	G_LED = 1;				//Hello Last
	MILLISECONDS (250);
	MILLISECONDS (250);
	G_LED = 0;				//Hello End

	//enable Serial Comms
	RCSTAbits.SPEN = 1;		// RX/TX Activated
	RCREG = 0;
	PIE1bits.RCIE = 1;		// Enables USART receive interrupt
	
	INTCONbits.GIE = 1;		// Enables all unmasked interrupts
	INTCONbits.PEIE = 1;	// Peripheral Interrupts enabled	


// Debugging: start immediately:
//	Key0Routine();

// Main Loop
	while(1) {
		WaitForKey();
		StatusSeqNumRead();
		if (gMov)
			UpdateSensors();
		if (RCSTAbits.OERR)
			RS232_Reset();
	}
}


/************* Control Algorithm *********************/
/* Runs at servo frequency (50Hz)
 *  be careful not to take too long!!
 */
void Control( void )
{
	long temp;
	long difftemp;
	unsigned char tempchar;
	
	// Open loop control: just copy gait to command:
	memcpy( gValueCmd, gValueServ, NUM_OF_SERV );
	
	temp = (long) gSensorValue[0] + (long) gSensorValue[1]; // + gSensorValue[2] + gSensorValue[3];
	temp = temp/2;	// average sensor value

	// move each servo towards average value
	difftemp = temp - (long) gSensorValue[0] + (long) gValueServ[0];
	if (difftemp > 255)
		difftemp = 255;
	if (difftemp < 0)
		difftemp = 0;	
	gValueCmd[0] = (unsigned char) difftemp;
	
	
	difftemp = temp - (long) gSensorValue[1] + (long) gValueServ[1];
	if (difftemp > 255)
		difftemp = 255;
	if (difftemp < 0)
		difftemp = 0;	
	gValueCmd[1] = (unsigned char) difftemp;

/* For scaling sensor range to full 8 bits - should really be done on the sensor board- JKK
	if (gSensorValue[0] < 72)
		gSensorValue[0] = 0;
	else
		gSensorValue[0] = gSensorValue[0] - 72;
	
	if (gSensorValue[0] > 36)
		gSensorValue[0] = 255;
	else
		gSensorValue[0] = gSensorValue[0]*7;
		
	gValueCmd[1] = gSensorValue[0];

*/

/*
	difftemp = (long) gSensorValue[2] - temp + (long) gValueServ[2];
	if (difftemp > 255)
		difftemp = 255;
	if (difftemp < 0)
		difftemp = 0;	
	gValueCmd[2] = (unsigned char) difftemp;

	difftemp = (long) gSensorValue[3] - temp + (long) gValueServ[3];
	if (difftemp > 255)
		difftemp = 255;
	if (difftemp < 0)
		difftemp = 0;	
	gValueCmd[3] = (unsigned char) difftemp;
*/

}

////////////////////////////////
// HIGH INTERRUPT SERVICE ROUTINE
////////////////////////////////

// Code Vectoring
#pragma code high_vector=0x08
void interrupt_at_high_vector(void)
{
_asm GOTO highISR _endasm
}
#pragma code /* return to the default code section */

// High ISR Code
#pragma interruptlow highISR
void highISR( void )
{
	mTempCounter = DelayCounter1;

	if (PIR1bits.TMR1IF)		// (1Tcy), did a Timer1 overflow occur?
		TMR1H = TMR1H | 0b10000000;	// (1Tcy) Yes, reload for a 1 second overflow

	if (PIR1bits.RCIF)
		Decode();

	if (PIR2bits.CCP2IF)
		TimerUpdate();

	DelayCounter1 = mTempCounter;

}

////////////////////////////////
// LOW INTERRUPT SERVICE ROUTINE
////////////////////////////////

// Code Vectoring
#pragma code low_vector=0x18
void interrupt_at_low_vector(void)
{
_asm GOTO highISR _endasm
}
#pragma code /* return to the default code section */

void TimerUpdate( void )	// must be less than 20ms (including interrupt overhead)
{
/* 	unsigned int max_step;
	unsigned int time_step;
*/
	PIR2bits.CCP2IF = 0;	// Clear compare interrupt flag
//	gFiftieths++;
	ServoUpdate();		// 2.09ms - 20902 instruction cycles
	if (gMov){
		if (gMaxStep) {
			gTimeStep--;
			if (!gTimeStep) {
				gStepCounter++;
				ReadStepFromFlash();
			}
		}
		Interpolate();		// Determine next command: 1.15ms (11511 instruction cycles - yikes!)
		Control();			// Modify based on controller: 20us (204 instruction cycles)
	}

//	if (gFiftieths == 50)
//			gFiftieths = 0;
//	
//	return;
}


void LoadTOC( void )
//LOAD TOC: FLASH --> RAM
{
	ReadFlash(FLASH_INDEX_OFFSET, gTOC, TOC_SIZE);
}

void SaveTOC( void )
{
	WriteFlash(FLASH_INDEX_OFFSET, gTOC, MAX_NUM_OF_SEQ*4+2);
}
	

void Key0Routine( void )
{
	TimerMovOn();
}

void Key0PressRoutine( void )
{
}

void Key1Routine( void )
{
	TimerMovOff();
}

// AUTO-ZERO SEQUENCE (UNCOMMENT IF REQUIRED)
void Key1PressRoutine( void )
{
	if (!gMov) {
		gGaitPosn = FLASH_OFFSET;
		ReadStepFromFlash();
		TimerOn();
		R_LED = 1;				// Key1PressRoutine Start (reset?)
		MILLISECONDS (250);
		MILLISECONDS (250);
		MILLISECONDS (250);
		MILLISECONDS (250);
		R_LED = 0;				// Key1Press Routine End (reset?)
		TimerOff();
	}
}



// Set default sequence
#pragma romdata gait_toc=FLASH_INDEX_OFFSET
const rom char initial_gait_toc[] = {
	0,	224,	// NUM OF SEQUENCE FLASH BYTES !!! FLASH_INDEX (TOC) NOT INCLUDED
	0,	8,		// SEQ 0 STEPS
	0,	0,		// SEQ 0 OFFSET
	0,	2,		// SEQ 1 STEPS
	0,	112,	// SEQ 1 OFFSET
	0,	2,		// SEQ 2 STEPS
	0,	140,	// SEQ 2 OFFSET
	0,	2,		// SEQ 3 STEPS
	0,	168,	// SEQ 3 OFFSET
	0,	2,		// SEQ 4 STEPS
	0,	196,	// SEQ 4 OFFSET
	0,	0,		// SEQ 5 STEPS
	0,	0,		// SEQ 5 OFFSET
	0,	0,		// SEQ 6 STEPS
	0,	0,		// SEQ 6 OFFSET
	0,	0,		// SEQ 7 STEPS
	0,	0,		// SEQ 7 OFFSET
	0,	0,		// SEQ 8 STEPS
	0,	0,		// SEQ 8 OFFSET
	0,	0,		// SEQ 9 STEPS
	0,	0,		// SEQ 9 OFFSET
	0,	0,		// SEQ 10 STEPS
	0,	0,		// SEQ 10 OFFSET
	0,	0,		// SEQ 11 STEPS
	0,	0,		// SEQ 11 OFFSET
	0,	0,		// SEQ 12 STEPS
	0,	0,		// SEQ 12 OFFSET
	0,	0,		// SEQ 13 STEPS
	0,	0,		// SEQ 13 OFFSET
	0,	0,		// SEQ 14 STEPS
	0,	0,		// SEQ 14 OFFSET
	0,	0,		// SEQ 15 STEPS
	0,	0		// SEQ 15 OFFSET
};
	
#pragma romdata gait=FLASH_OFFSET
const rom char initial_gait[] = {
	150, 128,		// SEQ 0 STEPS #0
	0,	0,
	0,	0,
	0,	0,
	0,	0,
	0,	0,
	0,	30,
	110, 128,		// SEQ 0 STEPS #1
	0,	0,
	0,	0,
	0,	0,
	0,	0,
	0,	10,
	0,	30,
	120, 128,		// SEQ 0 STEPS #2
	0,	0,
	0,	0,
	0,	0,
	0,	0,
	0,	20,
	0,	30,
	130, 128,		// SEQ 0 STEPS #3
	0,	0,
	0,	0,
	0,	0,
	0,	0,
	0,	30,
	0,	30,
	140, 128,		// SEQ 0 STEPS #4
	0,	0,
	0,	0,
	0,	0,
	0,	0,
	0,	40,
	0,	30,
	150, 128,		// SEQ 0 STEPS #5
	0,	0,
	0,	0,
	0,	0,
	0,	0,
	0,	50,
	0,	30,
	160, 128,		// SEQ 0 STEPS #6
	0,	0,
	0,	0,
	0,	0,
	0,	0,
	0,	60,
	0,	30,
	170, 128,		// SEQ 0 STEPS #7
	0,	0,
	0,	0,
	0,	0,
	0,	0,
	0,	70,
	0,	30,
	0,	0,	//<---	// SEQ 1 STEPS #0
	0,	0,
	0,	0,
	0,	0,
	0,	0,
	0,	0,
	0,	30,
	255,	0,	//<---	// SEQ 1 STEPS #1
	0,	0,
	0,	0,
	0,	0,
	0,	0,
	0,	0,
	0,	30,
	0,	0,	//<---	// SEQ 2 STEPS #0
	0,	0,
	0,	0,
	0,	0,
	0,	0,
	0,	0,
	0,	60,
	100,	0,	//<---	// SEQ 2 STEPS #1
	0,	0,
	0,	0,
	0,	0,
	0,	0,
	0,	0,
	0,	60,
	0,	0,	//<---	// SEQ 3 STEPS #0
	0,	0,
	0,	0,
	0,	0,
	0,	0,
	0,	0,
	0,	120,
	200,	0,	//<---	// SEQ 3 STEPS #1
	0,	0,
	0,	0,
	0,	0,
	0,	0,
	0,	0,
	0,	120,
	0,	0,	//<---	// SEQ 4 STEPS #0
	0,	0,
	0,	0,
	0,	0,
	0,	0,
	0,	0,
	0,	30,
	20,	0,	//<---	// SEQ 4 STEPS #1
	0,	0,
	0,	0,
	0,	0,
	0,	0,
	0,	0,
	0,	30
};
